from __future__ import annotations
from PySide6.QtCore import Signal, QObject, QEnum

from PySide6.QtQuick import QQuickItem
from typing import Optional, Union
from enum import Enum, auto
from PySide6.QtQml import QmlElement, QmlSingleton

from qtgql.codegen.py.runtime.queryhandler import SelectionConfig
from qtgql.tools import qproperty
from qtgql.codegen.py.runtime.bases import QGraphQListModel






{% for dep in context.dependencies %}
{{dep}}{% endfor %}


QML_IMPORT_NAME = "generated.{{context.config.env_name}}.types"
QML_IMPORT_MAJOR_VERSION = 1



{% for enum in context.enums %}
class {{enum.name}}(Enum):
    {% for member in enum.members %}
    {{member.name}} = auto()
    """{{member.description}}"""{% endfor %}

{% endfor %}

{% if context.enums %}
@QmlElement
class Enums(QObject):
    {% for enum in context.enums %}
    QEnum({{enum.name}})
    {% endfor %}
{% endif %}


class SCALARS:
    {% for scalar in context.custom_scalars %}
    {{scalar}} = {{scalar}}{% endfor %}

{% for type in context.types %}
class {{ type.name }}({{context.base_object_name}}):
    """{{  type.docstring  }}"""


    def __init__(self, parent: QObject = None, {% for f in type.fields %} {{f.name}}: Optional[{{f.annotation}}] = None, {% endfor %}):
        super().__init__(parent){% for f in type.fields %}
        self.{{  f.private_name  }} = {{f.name}} if {{f.name}} else {{f.default_value}}{% endfor %}

    def update(self, data, config: SelectionConfig) -> None:
        parent = self.parent()
        {%for f in type.fields %}
        if '{{f.name}}' in config.selections.keys():
            field_data = data.get('{{f.name}}', {{f.default_value}})
            {% if f.type.is_object_type %}
            if not field_data:
                self.{{f.setter_name}}(None)
            else:
                if self.{{f.private_name}} and self.{{f.private_name}}._id == field_data['id']:
                    self.{{f.private_name}}.update(field_data, config.selections['{{f.name}}'])
                else:
                    self.{{f.setter_name}}({{f.type.is_object_type.name}}.from_dict(
                        parent,
                        field_data,
                        config.selections['{{f.name}}']
                    ))
            {% elif f.type.is_model %}
            self.{{f.private_name}}.update(field_data, config.selections['{{f.name}}'])
            {% elif f.type.is_builtin_scalar %}
            if self.{{f.private_name}} != field_data:
                self.{{f.setter_name}}(field_data)
            {% elif f.is_custom_scalar %}
            new = SCALARS.{{f.is_custom_scalar.__name__}}.from_graphql(field_data)
            if new != self.{{f.private_name}}:
                self.{{f.setter_name}}(new)
            {% elif f.type.is_enum %}
            if self.{{f.private_name}}.name != field_data:
                self.{{f.setter_name}}({{f.type.is_enum.name}}[field_data])
            {% elif f.type.is_union() %}
            type_name = field_data['__typename']
            choice = config.selections['{{f.name}}'].choices[type_name]
            if self.{{f.private_name}} and self.{{f.private_name}}._id == field_data['id']:
                self.{{f.private_name}}.update(field_data, choice)
            else:
                self.{{f.setter_name}}(self.type_map[type_name].from_dict(parent, field_data, choice))
            {% endif %}
            {% endfor %}

    @classmethod
    def from_dict(cls, parent, data: dict, config: SelectionConfig) -> {{type.name}}:
        {% if type.id_is_optional %}
        if id_ := data.get('id', None):
            if instance := cls.__store__.get_node(id_):
                instance.update(data, config)
                return instance
        {% elif type.has_is_field %}
        if instance := cls.__store__.get_node(data['id']):
            instance.update(data, config)
            return instance
        {% endif %}
        {% if type.has_is_field%}
        else:
        {% endif %}
            inst = cls(parent=parent)
            {% for f in type.fields %}
            if '{{f.name}}' in config.selections.keys():
                field_data = data.get('{{f.name}}', {{f.default_value}})
                {% if f.type.is_object_type %}
                if field_data:
                    inst.{{f.private_name}} = {{f.type.is_object_type.name}}.from_dict(
                        parent,
                        field_data,
                        config.selections['{{f.name}}']
                    )
                {% elif f.type.is_model %}
                node_config = config.selections['{{f.name}}']
                inst.{{f.private_name}} = QGraphQListModel(
                    parent=parent,
                    data=[{{f.type.is_model.name}}.from_dict(parent, data=node, config=node_config) for node in field_data],
                    default_type={{f.type.is_model.name}},
                )
                {% elif f.type.is_builtin_scalar %}
                inst.{{f.private_name}} = field_data
                {% elif f.is_custom_scalar %}
                inst.{{f.private_name}} = SCALARS.{{f.is_custom_scalar.__name__}}.from_graphql(field_data)
                {% elif f.type.is_enum %}
                inst.{{f.private_name}} = {{f.type.is_enum.name}}[field_data]
                {% elif f.type.is_union() %}
                type_name = field_data['__typename']
                choice = config.selections['{{f.name}}'].choices[type_name]
                inst.{{f.private_name}} = cls.type_map[type_name].from_dict(parent, field_data, choice)
                {% endif %}
                {% endfor %}
            {% if type.id_is_optional %}
            if inst.id:
                cls.__store__.set_node(inst)
            {% elif type.has_id_field and not type.id_is_optional %}
            cls.__store__.set_node(inst)
            {% endif %}
            return inst

    {% for f in type.fields %}
    {{ f.signal_name }} = Signal()

    def {{ f.setter_name }}(self, v: {{  f.annotation  }}) -> None:
        self.{{  f.private_name  }} = v
        self.{{  f.signal_name  }}.emit()

    @qproperty(type={{f.property_type}}, fset={{ f.setter_name }}, notify={{f.signal_name}})
    def {{ f.name }}(self) -> {{ f.fget_annotation }}:
        {{f.fget}}
    {% endfor %}

{% endfor %}






from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Generic, Optional, TypeVar

from PySide6.QtCore import QAbstractListModel, QByteArray, QObject, Qt, Signal

if TYPE_CHECKING:
    from typing_extensions import Self

    from qtgql.codegen.py.runtime.queryhandler import SelectionConfig
from qtgql.tools import qproperty, slot

__all__ = ["QGraphQListModel", "get_base_graphql_object"]


class _BaseQGraphQLObject(QObject):
    type_map: dict[str, type[_BaseQGraphQLObject]]

    id: str
    __singleton__: Self
    __store__: ClassVar[QGraphQLObjectStore[Self]]

    def __init_subclass__(cls, **kwargs):
        cls.type_map[cls.__name__] = cls  # required to instantiate unions.
        cls.__store__ = QGraphQLObjectStore()

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)

    @qproperty(str, constant=True)
    def typename(self) -> str:
        return self.__class__.__name__

    @classmethod
    def from_dict(cls, parent: QObject, data: dict, config: SelectionConfig):
        raise NotImplementedError

    def update(self, data, config: SelectionConfig) -> None:
        raise NotImplementedError

    @classmethod
    def default_instance(cls) -> Self:
        # used for default values.
        try:
            return cls.__singleton__  # type: ignore
        except AttributeError:
            cls.__singleton__ = cls()
            return cls.__singleton__


T_BaseQGraphQLObject = TypeVar("T_BaseQGraphQLObject", bound=_BaseQGraphQLObject)


class QGraphQLObjectStore(Generic[T_BaseQGraphQLObject]):
    def __init__(self) -> None:
        self._data: dict[str, T_BaseQGraphQLObject] = {}

    def get_node(self, id_: str) -> Optional[T_BaseQGraphQLObject]:
        assert id_
        return self._data.get(id_, None)

    def set_node(self, node: T_BaseQGraphQLObject):
        assert node.id
        self._data[node.id] = node


class QGraphQListModel(QAbstractListModel, Generic[T_BaseQGraphQLObject]):
    OBJECT_ROLE = Qt.ItemDataRole.UserRole + 1
    _role_names = {OBJECT_ROLE: QByteArray("object")}  # type: ignore
    currentIndexChanged = Signal()

    def __init__(
        self,
        parent: Optional[QObject],
        default_type: T_BaseQGraphQLObject,
        data: list[T_BaseQGraphQLObject],
    ):
        super().__init__(parent)

        self._data = data
        self.default_type = default_type
        self._default_object = default_type.default_instance()
        self._current_index: int = 0

    @slot
    def set_current_index(self, i: int) -> None:
        self._current_index = i
        self.currentIndexChanged.emit()

    @qproperty(int, notify=currentIndexChanged, fset=set_current_index)
    def currentIndex(self) -> int:
        return self._current_index

    @qproperty(QObject, notify=currentIndexChanged)  # type: ignore
    def currentObject(self) -> Optional[T_BaseQGraphQLObject]:
        try:
            return self._data[self._current_index]
        except IndexError:
            return self._default_object

    def rowCount(self, *args, **kwargs) -> int:
        return len(self._data)

    def roleNames(self) -> dict:
        return self._role_names  # type: ignore

    def data(self, index, role=...) -> Optional[T_BaseQGraphQLObject]:
        if index.row() < len(self._data) and index.isValid():
            if role == self.OBJECT_ROLE:
                return self._data[index.row()]
            raise NotImplementedError(
                f"role {role} is not a valid role for {self.__class__.__name__}"
            )

    def append(self, node: T_BaseQGraphQLObject) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()

    @slot
    def insert(self, index: int, v: T_BaseQGraphQLObject):
        model_index = self.index(index)
        if index <= self.rowCount() + 1:
            self.beginInsertRows(model_index, index, index)
            self._data.insert(index, v)
            self.endInsertRows()

    def update(self, data: list[dict], node_selection: SelectionConfig) -> None:
        new_len = len(data)
        prev_len = self.rowCount()
        if new_len < prev_len:
            # crop the list to the arrived data length.
            self.removeRows(new_len, prev_len - new_len)
        for index, node in enumerate(data):
            id_ = node.get("id", None)
            if id_ and self._data[index].id == id_:
                # same node on that index just call update there is no need call model signals.
                self._data[index].update(data[index], node_selection)
            else:
                # get or create node if wasn't on the correct index.
                # Note: it is safe to call [].insert(50, 50) (although index 50 doesn't exist).
                self.insert(index, self.default_type.from_dict(self, data[index], node_selection))

    def removeRows(self, row: int, count: int, parent=None) -> bool:
        if row + count <= self.rowCount():
            self.beginRemoveRows(self.index(0).parent(), row, count)
            end = self._data[row + count :]
            start = self._data[:row]
            self._data = start + end
            self.endRemoveRows()
            return True
        return False


def get_base_graphql_object(name: str) -> type[_BaseQGraphQLObject]:
    """
    :param name: valid attribute name (used by codegen to import it).
    :returns: A type to be extended by all generated types.
    """
    type_map: dict[str, type[_BaseQGraphQLObject]] = {}
    return type(
        name, (_BaseQGraphQLObject,), {"type_map": type_map, "__store__": QGraphQLObjectStore()}
    )  # type: ignore


BaseGraphQLObject = get_base_graphql_object("BaseGraphQLObject")

T_BaseModel = TypeVar("T_BaseModel", bound=QGraphQListModel)



from typing import Optional, Union
from PySide6.QtCore import Signal, QObject
from PySide6.QtQml import QmlElement, QmlSingleton
from qtgql.codegen.py.runtime.queryhandler import BaseQueryHandler, UseQueryABC, SelectionConfig
from qtgql.gqltransport.client import  GqlClientMessage, QueryPayload
from qtgql.codegen.py.runtime.bases import QGraphQListModel
from objecttypes import * # noqa


QML_IMPORT_NAME = "generated.{{context.config.env_name}}"
QML_IMPORT_MAJOR_VERSION = 1






{% for query in context.queries %}
@QmlElement
@QmlSingleton
class {{query.name}}(BaseQueryHandler[{{query.field.annotation}}]):
    ENV_NAME = "{{context.config.env_name}}"
    _message_template = GqlClientMessage(payload=QueryPayload(query="""{{query.query}}""", operationName="{{query.name}}"))

    OPERATION_CONFIG = {{query.operation_config}}

    def on_data(self, message: dict) -> None:
        config = self.OPERATION_CONFIG
        # TODO: this should be a copy of the `from_dict()` template.
        field_data = message.get('{{query.field.name}}', None)
        {% if query.field.type.is_object_type %}
        if not field_data:
            emit = False
            if self._data:
                emit = True
            self._data = None
            if emit:
                self.dataChanged.emit()
        else:
            if self._data and self._data._id == field_data['id']:
                self._data.update(field_data, config)
            else:
                self._data = {{query.field.type.is_object_type.name}}.from_dict(
                    self,
                    field_data,
                    config
                )
                self.dataChanged.emit()
        {% elif query.field.type.is_model %}
        if field_data:
            if self._data:
                self._data.update(field_data, config)
            else:
                self._data = QGraphQListModel(
                    parent=self,
                    data=[{{ query.field.type.is_model.name }}.from_dict(self, data=node, config=config) for node in field_data],
                    default_type={{ query.field.type.is_model.name }},
                )
                self.dataChanged.emit()
        {% endif %}

{% endfor %}


def init() -> None:
    {% for query in context.queries %}
    {{query.name}}()
    {% endfor %}
@QmlElement
class UseQuery(UseQueryABC):
    ENV_NAME = "{{context.config.env_name}}"